/*
 * Copyright 2007 The Apache Software Foundation.
 *
 * Licensed under the Apache License, Version 2.0 (the "License")
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package net.sf.hibernate4gwt.core.beanlib.merge;

import java.lang.reflect.Method;

import net.sf.beanlib.spi.DetailedBeanPopulatable;
import net.sf.hibernate4gwt.core.IPersistenceUtil;
import net.sf.hibernate4gwt.core.beanlib.BeanlibThreadLocal;
import net.sf.hibernate4gwt.exception.CloneException;
import net.sf.hibernate4gwt.pojo.ILazyPojo;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

/**
 * Populatable for POJO with the lazy information of the Hibernate POJO
 * This populatable is used to fill back an Hibernate POJO from a pure clone
 * POJO (not implementing ILazyPojo).
 * Only the not lazy properties of the Hibernate POJO are filled.
 * To be correct, the clone and merge operations must be managed with the
 * <b>same</b> Hibernate POJO, or a POJO with the <b>same</b> fetching strategy
 * @author bruno.marchesson
 */
public class MergeBeanPopulatable implements DetailedBeanPopulatable 
{
	//----
	// Attributes
	//----
	/**
	 * Log channel
	 */
	private static Log _log = LogFactory.getLog(MergeBeanPopulatable.class);
	
	/**
	 * The associated persistence utils
	 */
	private IPersistenceUtil _persistenceUtil;
	
	//----
	// Properties
	//----
	/**
	 * @return the persistence Util implementation to use
	 */
	public IPersistenceUtil getPersistenceUtil()
	{
		return _persistenceUtil;
	}

	/**
	 * @param util the persistenceUtil to set
	 */
	public void setPersistenceUtil(IPersistenceUtil util)
	{
		_persistenceUtil = util;
	}
	
	//-------------------------------------------------------------------------
	//
	// Constructor
	//
	//-------------------------------------------------------------------------
	/**
	 * Constructor
	 */
	public MergeBeanPopulatable(IPersistenceUtil persistenceUtil)
	{
		setPersistenceUtil(persistenceUtil);
	}
	
	//-------------------------------------------------------------------------
	//
	// DetailedBeanPopulatable implementation
	//
	//-------------------------------------------------------------------------
	/* (non-Javadoc)
	 * @see net.sf.beanlib.spi.DetailedBeanPopulatable#shouldPopulate(java.lang.String, java.lang.Object, java.lang.reflect.Method, java.lang.Object, java.lang.reflect.Method)
	 */
	public boolean shouldPopulate(String propertyName, 
								  Object fromBean, 
								  Method readerMethod, 
								  Object toBean, 
								  Method setterMethod)
	{
	//	Is the property lazy loaded ?
	//
		try
		{
			if (fromBean instanceof ILazyPojo == true)
			{
			//	Check if merged value is null
			//
				Object fromValue = readerMethod.invoke(fromBean, new Object[] {});
				
			//	ILazyPojo : just check the lazy properties string
			//
				if ((fromValue == null) && 
					(((ILazyPojo) fromBean).isLazyProperty(propertyName) == true))
				{
					_log.info(fromBean.toString()+ "." + propertyName + " --> lazy loaded ");
					BeanlibThreadLocal.setDestinationBean(null);
					return false;
				}
				else
				{
					if (fromBean.getClass().isAssignableFrom(toBean.getClass()))
					{
					//	Set the proxy as destination bean (no new instance needed for merge)
					//
						BeanlibThreadLocal.setDestinationBean(readerMethod.invoke(toBean, new Object[] {}));
					}
					else
					{
					//	Create a new reader method
					//
						Method toReaderMethod = toBean.getClass().getMethod(readerMethod.getName(), 
																			readerMethod.getParameterTypes());
						BeanlibThreadLocal.setDestinationBean(toReaderMethod.invoke(toBean, new Object[] {}));
					}
					return true;
				}
			}
			else if (fromBean.getClass().isAssignableFrom(toBean.getClass()) == false)
			{
			//  Cannot use the Hibernate bean with a DTO class (Java5 DTO class) !
			//
				throw new CloneException("Explicit DTO (Java5) must be declared as extending LazyGwtPojo !");
			}
			else
			{
			//	Check if merged value is null
			//
				Object fromValue = readerMethod.invoke(fromBean, new Object[] {});
				
			//	Call the readerMethod on the Hibernate Bean to determine the lazy property
			//
				Object toValue = readerMethod.invoke(toBean, new Object[] {});
				if ((fromValue == null) &&
					(_persistenceUtil.isProxyInitialized(toValue) == false))
				{
					_log.info(fromBean.toString()+ "." + propertyName + " --> lazy loaded ");
					BeanlibThreadLocal.setDestinationBean(null);
					return false;
				}
				else
				{
				//	Set the proxy as destination bean (no new instance needed for merge)
				//
					BeanlibThreadLocal.setDestinationBean(toValue);
					return true;
				}
			}
		}
		catch(Exception ex)
		{
			throw new RuntimeException(ex);
		}
	}

}
